
import cffi, re

# Our defs we want to use
cdef_evtapi = r"""
typedef enum { EvtRpcLogin = 1 } EVT_LOGIN_CLASS;

typedef HANDLE EVT_HANDLE;

EVT_HANDLE WINAPI EvtOpenSession(
    _In_ EVT_LOGIN_CLASS LoginClass,
    _In_ PVOID           Login,
    _In_ DWORD           Timeout,
    _In_ DWORD           Flags
);

typedef struct _EVT_RPC_LOGIN {
  LPWSTR Server;
  LPWSTR User;
  LPWSTR Domain;
  LPWSTR Password;
  DWORD  Flags;
} EVT_RPC_LOGIN;

typedef enum _EVT_RPC_LOGIN_FLAGS { 
  EvtRpcLoginAuthDefault    = 0,
  EvtRpcLoginAuthNegotiate  = 1,
  EvtRpcLoginAuthKerberos   = 2,
  EvtRpcLoginAuthNTLM       = 3
} EVT_RPC_LOGIN_FLAGS;

typedef enum _EVT_SUBSCRIBE_NOTIFY_ACTION { 
  EvtSubscribeActionError    = 0,
  EvtSubscribeActionDeliver  = 1
} EVT_SUBSCRIBE_NOTIFY_ACTION;

typedef DWORD ( WINAPI *EVT_SUBSCRIBE_CALLBACK)(
   EVT_SUBSCRIBE_NOTIFY_ACTION Action,
   PVOID                       UserContext,
   EVT_HANDLE                  Event
);

EVT_HANDLE WINAPI EvtSubscribe(
  _In_ EVT_HANDLE             Session,
  _In_ HANDLE                 SignalEvent,
  _In_ LPCWSTR                ChannelPath,
  _In_ LPCWSTR                Query,
  _In_ EVT_HANDLE             Bookmark,
  _In_ PVOID                  context,
  _In_ EVT_SUBSCRIBE_CALLBACK Callback,
  _In_ DWORD                  Flags
);

EVT_HANDLE WINAPI EvtQuery(
  _In_ EVT_HANDLE Session,
  _In_ LPCWSTR    Path,
  _In_ LPCWSTR    Query,
  _In_ DWORD      Flags
);

typedef enum _EVT_QUERY_FLAGS { 
  EvtQueryChannelPath          = 0x1,
  EvtQueryFilePath             = 0x2,
  EvtQueryForwardDirection     = 0x100,
  EvtQueryReverseDirection     = 0x200,
  EvtQueryTolerateQueryErrors  = 0x1000
} EVT_QUERY_FLAGS;

BOOL WINAPI EvtNext(
  _In_  EVT_HANDLE  ResultSet,
  _In_  DWORD       EventArraySize,
  _In_  EVT_HANDLE* EventArray,
  _In_  DWORD       Timeout,
  _In_  DWORD       Flags,
  _Out_ PDWORD      Returned
);

typedef enum _EVT_RENDER_FLAGS { 
  EvtRenderEventValues  = 0,
  EvtRenderEventXml     = 1,
  EvtRenderBookmark     = 2
} EVT_RENDER_FLAGS;

BOOL WINAPI EvtRender(
  _In_  EVT_HANDLE Context,
  _In_  EVT_HANDLE Fragment,
  _In_  DWORD      Flags,
  _In_  DWORD      BufferSize,
  _In_  PVOID      Buffer,
  _Out_ PDWORD     BufferUsed,
  _Out_ PDWORD     PropertyCount
);

BOOL WINAPI EvtClose(
  _In_ EVT_HANDLE Object
);
                
typedef enum _EVT_SUBSCRIBE_FLAGS { 
  EvtSubscribeToFutureEvents       = 1,
  EvtSubscribeStartAtOldestRecord  = 2,
  EvtSubscribeStartAfterBookmark   = 3,
  EvtSubscribeOriginMask           = 0x3,
  EvtSubscribeTolerateQueryErrors  = 0x1000,
  EvtSubscribeStrict               = 0x10000
} EVT_SUBSCRIBE_FLAGS;

extern "Python" DWORD WINAPI SubscriptionCallback(EVT_SUBSCRIBE_NOTIFY_ACTION, PVOID, EVT_HANDLE);
"""

cdef_kernel32 = r"""
DWORD WINAPI GetLastError(void);
"""

cdef = cdef_evtapi + cdef_kernel32

# Simple source file
source = r"""
#include <windows.h>
#include <WinEvt.h>

void main()
{
    return 0;
}
"""

def ffibuilder(name="_winevt"):
    """Do out-of-line build."""

    # Insantiate a class
    ffi = cffi.FFI()

    # Setup our definitions
    ffi.cdef(re.sub(r"\b(_In_|_Inout_|_Out_|_Outptr_)(opt_)?\b", " ",cdef))

    # Setup our binary
    ffi.set_source(name,source,libraries=["Wevtapi"])

    return ffi


# Default build and compile it.
if __name__ == "__main__":
    ffi = ffibuilder()

    # Compile it!
    ffi.compile(verbose=True)




""" Class for subscribing to Windows Event Logs. """

import logging
logger = logging.getLogger("EventLog.Subscribe")

class Subscribe:

    def __init__(self, path, query, callback, start_from=None, tolerate_query_errors = False, strict = False):
        """
        path = The name of the channel or the full path to a log file that contains the events that you want to query. You can specify an .evt, .evtx, or.etl log file. The path is required if the Query parameter contains an XPath query; the path is ignored if the Query parameter contains a structured XML query and the query specifies the path.
        query = A query that specifies the types of events that you want to retrieve. You can specify an XPath 1.0 query or structured XML query. If your XPath contains more than 20 expressions, use a structured XML query. To receive all events, set this parameter to None or "*".
        callback = python function to accept callbacks from the event subscription.
        start_from = Where to start subscription from. Default is "future". See property.
        tolerate_query_errors = Should we try to push through errors in the query. Defaults to False.
        strict = Should we be strict about our query. See Property for details.
        """

        self.path = path
        self.query = query
        self.start_from = start_from or "future"
        self.tolerate_query_errors = tolerate_query_errors
        self.strict = strict
        self.callback = callback

    def __repr__(self):
        return "<Subscribe path=\"{0}\" query=\"{1}\" callback=\"{2}\">".format(self.path, self.query, self.__callback_python.__name__)

    ##############
    # Properties #
    ##############

    @property
    def strict(self):
        """Forces the EvtSubscribe call to fail if you specify EvtSubscribeStartAfterBookmark and the bookmarked event is not found (the return value is ERROR_NOT_FOUND). Also, set this flag if you want to receive notification in your callback when event records are missing."""

        return self.__strict

    @strict.setter
    def strict(self, strict):

        if type(strict) is not bool:
            raise Exception("Invalid type for strict of {0}".format(type(strict)))

        self.__strict = strict

    @property
    def tolerate_query_errors(self):
        """Complete the subscription even if the part of the query generates an error (is not well formed). The service validates the syntax of the XPath query to determine if it is well formed. If the validation fails, the service parses the XPath into individual expressions. It builds a new XPath beginning with the left most expression. The service validates the expression and if it is valid, the service adds the next expression to the XPath. The service repeats this process until it finds the expression that is failing. It then uses the valid expressions that it found beginning with the leftmost expression as the XPath query (which means that you may not get the events that you expected). If no part of the XPath is valid, the EvtSubscribe call fails."""

        return self.__tolerate_query_errors

    @tolerate_query_errors.setter
    def tolerate_query_errors(self, tolerate_query_errors):
        if type(tolerate_query_errors) is not bool:
            raise Exception("Invalid type for tolerate_query_errors of {0}".format(type(tolerate_query_errors)))

        self.__tolerate_query_errors = tolerate_query_errors

    @property
    def flags(self):
        """ Flags for this subscription. """

        flags = 0

        if self.start_from == "future":
            flags |= evtapi.EvtSubscribeToFutureEvents

        elif self.start_from == "oldest":
            flags |= evtapi.EvtSubscribeStartAtOldestRecord

        if self.tolerate_query_errors:
            flags |= evtapi.EvtSubscribeTolerateQueryErrors

        if self.strict:
            flags |= evtapi.EvtSubscribeStrict

        return flags

    @property
    def start_from(self):
        """Where to start our subscription from. Options are:
        future -- Only get future events, no events already in the log.
        oldest -- Get all records that are valid for the query, and then subscribe to all future events.
        bookmark -- Use a bookmark. Not yet implemented
        """

        return self.__start_from

    @start_from.setter
    def start_from(self, start_from):
        if type(start_from) is not str:
            raise Exception("Attempting to set invalid start_from type of {0}".format(type(start_from)))

        start_from = start_from.lower()
        
        if start_from not in ["future", "oldest"]:
            raise Exception("start_from of \"{0}\" not valid or not supported yet.".format(start_from))

        self.__start_from = start_from


    @property
    def callback(self):
        return self.__callback

    @callback.setter
    def callback(self, callback):
        if type(callback) is not type(lambda x: x):
            raise Exception("Did not receive function for callback. Recieved invalid type {0}".format(type(callback)))

        if len(signature(callback).parameters) != 3:
            raise Exception("Python callback function must accept 3 parameters: action, pContext, Event")

        # Keep track of the original function
        self.__callback_python = callback
        
        ################
        # Add the hook #
        ################

        if out_of_line:
            # out-of-line is preferred
            @ffi.def_extern()
            def SubscriptionCallback(action, pContext, hEvent):
                # TODO: Watch action for error
                # https://msdn.microsoft.com/en-us/library/windows/desktop/aa385596(v=vs.85).aspx
                # 
                callback(action, pContext, Event(hEvent, close=False))

                # TODO: Do we always assume success?
                return 1

            self.__callback = SubscriptionCallback
            cb_ptr = evtapi.SubscriptionCallback

        else:
            # In-line callback setup
            @ffi.callback("DWORD WINAPI SubscriptionCallback(EVT_SUBSCRIBE_NOTIFY_ACTION, PVOID, EVT_HANDLE)")
            def SubscriptionCallback(action, pContext, hEvent):
                callback(action, pContext, Event(hEvent, close=False))

                # TODO: Do we always assume success?
                return 1

            self.__callback = SubscriptionCallback
            cb_ptr = SubscriptionCallback

        #
        # Subscribe to the events
        #

        ret = evtapi.EvtSubscribe(ffi.NULL, ffi.NULL, self.path, self.query, ffi.NULL, ffi.NULL, cb_ptr, self.flags)

        if not ret:
            logger.error("Something went wrong subscribing...")

    @property
    def path(self):
        """The name of the channel or the full path to a log file that contains the events that you want to query. You can specify an .evt, .evtx, or.etl log file. The path is required if the Query parameter contains an XPath query; the path is ignored if the Query parameter contains a structured XML query and the query specifies the path."""

        return self.__path

    @path.setter
    def path(self, path):
        # TODO: Error and type checking
        self.__path = path

    @property
    def query(self):
        """A query that specifies the types of events that you want to retrieve. You can specify an XPath 1.0 query or structured XML query. If your XPath contains more than 20 expressions, use a structured XML query. To receive all events, set this parameter to NULL or "*" """
        return self.__query

    @query.setter
    def query(self, query):
        # TODO: Error and type checking
        self.__query = query

from inspect import signature
from .. import ffi, evtapi, out_of_line
from winevt.EventLog.Event import Event



""" Defines an Event Log Query """

class Query:

    def __init__(self, path, query = None, direction = None):
        """
        path = The name of the channel or the full path to a log file that contains the events that you want to query. You can specify an .evt, .evtx, or.etl log file. The path is required if the Query parameter contains an XPath query; the path is ignored if the Query parameter contains a structured XML query and the query specifies the path.
        query = A query that specifies the types of events that you want to retrieve. You can specify an XPath 1.0 query or structured XML query. If your XPath contains more than 20 expressions, use a structured XML query. To receive all events, set this parameter to None or "*".
        direction = query "forward" or "backward" from your search.
        """

        # TODO: Allow credential passing
        # TODO: Allow remote querying

        self.path = path
        self.query = query or "*"
        self.direction = direction

        # Grab a handle to this query
        self.handle = evtapi.EvtQuery(ffi.NULL, self.path, self.query, self.flags)

    def __del__(self):
        # Be sure to clean up our query
        try:
            evtapi.EvtClose(self.handle)
        except:
            pass

    def __repr__(self):
        return "<Query path={0} query={1}>".format(self.path, self.query)

    # We'll handle the iterator ourself
    def __iter__(self):
        return self

    def __next__(self):
        # Loop until we get no more events
        # TODO: Check for errors on fail

        # Creating new objects each time
        evt_array = ffi.new("EVT_HANDLE *")
        ret = ffi.new("PDWORD")

        if not evtapi.EvtNext(self.handle, 1, evt_array, 60, 0, ret):
            # TODO: Check for errors (GetLastError())
            raise StopIteration

        return Event(ffi.unpack(evt_array, 1)[0])


    ##############
    # Properties #
    ##############

    @property
    def flags(self):
        """ Flags to be used for this query. """
        flags = 0
        
        # Add forward and back
        if self.direction == "forward":
            flags |= evtapi.EvtQueryForwardDirection

        else:
            flags |= evtapi.EvtQueryReverseDirection

        # Determine if we're talking about a file
        if os.path.isfile(self.path):
            flags |= evtapi.EvtQueryFilePath

        else:
            flags |= evtapi.EvtQueryChannelPath

        return flags


    @property
    def path(self):
        """The name of the channel or the full path to a log file that contains the events that you want to query. You can specify an .evt, .evtx, or.etl log file. The path is required if the Query parameter contains an XPath query; the path is ignored if the Query parameter contains a structured XML query and the query specifies the path."""

        return self.__path

    @path.setter
    def path(self, path):
        # TODO: Error and type checking
        self.__path = path

    @property
    def query(self):
        """A query that specifies the types of events that you want to retrieve. You can specify an XPath 1.0 query or structured XML query. If your XPath contains more than 20 expressions, use a structured XML query. To receive all events, set this parameter to NULL or "*" """
        return self.__query

    @query.setter
    def query(self, query):
        # TODO: Error and type checking
        self.__query = query

    @property
    def direction(self):
        """The direction to move from your search. Options are "forward" (default) and "backward"."""
        return self.__direction

    @direction.setter
    def direction(self, direction):

        if type(direction) not in [type(None), str]:
            raise Exception("Invalid direction of type {0}".format(type(direction)))
        
        # Normalize it
        if direction is None:
            direction = "forward"

        else:
            direction = direction.lower()

        if direction not in ["forward","backward"]:
            raise Exception("Invalid direction recieved. Valid options are \"forward\" and \"backward\"")

        self.__direction = direction

import os
from .. import ffi, evtapi
from winevt.EventLog.Event import Event



import logging

logging.basicConfig(level=logging.WARNING)
logger = logging.getLogger("setup.py")


# Always prefer setuptools over distutils
from setuptools import setup, find_packages
# To use a consistent encoding
from codecs import open
import os
from setuptools.command.install import install
from setuptools.command.build_py import build_py
from setuptools.command.sdist import sdist
import shutil
from glob import glob
import sys

here = os.path.abspath(os.path.dirname(__file__))

def _build_ffi():
    # This dance is hackish. Probably better way to do this.

    # chdir so we know where we're importing from
    old_dir = os.path.abspath(os.curdir)
    os.chdir(os.path.join(here,"winevt"))

    # Apprently sdist doesn't have "." added by default?
    sys.path.append(".")

    # Try compile it, but be OK with failure
    try:
        # Import our ffi builder
        from winevt_build import ffibuilder

        ffibuilder().compile()
        shutil.copyfile(glob("_winevt.*.pyd")[0],"_winevt.pyd")
    except Exception as e:
        raise Exception(e)
        pass

    # Put us back in our original directory
    os.chdir(old_dir)


class CustomBuildPyCommand(build_py):
    """ Handle generating pyd file but not erroring if we can't. """
    def run(self):
        self.execute(_build_ffi, (), msg='Building ffi')
        build_py.run(self)

class CustomSdistCommand(sdist):
    """ Make sure we generate a new pyd when creating our sdist. """
    def run(self):
        self.execute(_build_ffi, (), msg='Building ffi')
        sdist.run(self)


# Get the long description from the README file
#with open(path.join(here, 'README.md'), encoding='utf-8') as f:
#    long_description = f.read()
long_description = "See website for more info."

setup(
    name='winevt',
    version='0.0.2',
    description='Script to programmatically interface with Windows Events.',
    long_description=long_description,
    url='https://github.com/owlz/winevt',
    author='Michael Bann',
    author_email='self@bannsecurity.com',
    license='MIT',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: End Users/Desktop',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Operating System :: Microsoft :: Windows',
        'Environment :: Console'
    ],
    keywords='windows event evt evtx',
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),
    install_requires=["cffi>=1.0.0","untangle"],
    setup_requires=["cffi>=1.0.0"],
    #cffi_modules=["winevt/winevt_build.py:ffibuilder"],
    extras_require={
        'dev': ['ipython'],
    },
    cmdclass={
        #'install': CustomInstallCommand,
        'build_py': CustomBuildPyCommand,
        'sdist': CustomSdistCommand,
    },
    package_data={'winevt': ['_winevt.pyd']},
)




import logging

#logging.basicConfig(level=logging.INFO)
logging.basicConfig(level=logging.WARNING)
logger = logging.getLogger("winevt")

#
# Load up whichever way we can
#

# Assume it's in-line unless we get otherwise
out_of_line = False

try:
    from ._winevt import ffi, lib as evtapi
    # Loading inline, these will be the same
    kernel32 = evtapi
    logger.info("Loaded in-line, user compiled evtapi")
    out_of_line = True

except ModuleNotFoundError:
    logger.warn("Looks like you didn't successfully compile your own out-of-line pyd. Falling back to in-ine mode. This is going to be less efficient and it's recommended you compile your own. To fix this, do the following:\n    1) Check this page and install the correct compiler for your version of python: https://blogs.msdn.microsoft.com/pythonengineering/2016/04/11/unable-to-find-vcvarsall-bat/\n    2) Re-install winevt (pip install -U winevt)")
    # In-line mode
    from .winevt_build import ffibuilder
    ffi = ffibuilder()
    evtapi = ffi.dlopen("Wevtapi.dll")
    kernel32 = ffi.dlopen("Kernel32.dll")




""" Define an Event """

LogLevels = {
        0: 'Undefined',
        1: 'Critical',
        2: 'Error',
        3: 'Warning',
        4: 'Information',
        5: 'Verbose'
}

class Event:

    def __init__(self, handle, max_buf_size=None, close=None):
        """
        handle = event handle (such as returned by Query)
        max_buf_size = optionally set the max buffer size for returning things. Defaults to 65536
        close = boolean, should we automatically close this event. Microsoft is quirky. Default True
        """

        self.__xml = None
        self.handle = handle
        self._max_buf_size = max_buf_size or 65536
        self.close = close or True

        self.xml # Populate the event

    def __repr__(self):
        return "<Event EventID={0} Level={1}>".format(self.EventID, self.LevelStr)


    def __del__(self):
        # Sometimes we don't close it ourselves
        if not self.close:
            return

        # Be sure to clean up our event
        try:
            evtapi.EvtClose(self.handle)
        except:
            pass

    ##############
    # Properties #
    ##############

    @property
    def close(self):
        """ Should we close this event on distruction of the class? """
        return self.__close

    @close.setter
    def close(self, close):
        if type(close) is not bool:
            raise Exception("Attempting to set close value to non-bool of {0}".format(type(close)))

        self.__close = close

    @property
    def xml(self):
        """ Returns the full XML dump of the event. """

        # Basically cache it
        if self.__xml != None:
            return self.__xml

        # Create some vars
        ret = ffi.new("PDWORD")
        ret2 = ffi.new("PDWORD")
        buf = ffi.new("PVOID[{0}]".format(int(self._max_buf_size/8)))

        if not evtapi.EvtRender(ffi.NULL, self.handle, evtapi.EvtRenderEventXml, self._max_buf_size, buf, ret, ret2):
            raise Exception("Something failed when trying to render...")

        # Save the buf
        self.__xml = ffi.string(ffi.cast("wchar_t *",buf))

        # Untangle and save as attributes
        dom = untangle.parse(self.__xml)
        for child in dom.Event.children:
            if not hasattr(self, child._name):
                setattr(self, child._name, child)

        # Return the buf
        return self.__xml
    
    @property
    def handle(self):
        return self.__handle

    @handle.setter
    def handle(self, handle):
        
        if type(handle) is not ffi.CData:
            raise Exception("handle must be of type ffi.CData")

        self.__handle = handle

    @property
    def EventID(self):
        # Helper function. Not strictly necessary as you can parse the dom or XML
        return int(self.System.EventID.cdata)

    @property
    def Level(self):
        return int(self.System.Level.cdata)

    @property
    def LevelStr(self):
        return LogLevels[self.Level]


#from _winevt import ffi, lib as evtapi
from .. import ffi, evtapi
import untangle
